{-# language PatternSynonyms #-}
module TypesWorkshop.Generics.Data where

import Data.Typeable (Typeable)
import GHC.Generics (Generic)
import qualified Unbound.Generics.LocallyNameless as U
import qualified Unbound.Generics.Unify as U

type Name = String

data Declaration = Declaration { declName :: Name, declType :: TypeScheme, declBody :: Term }

type TypeVariable = U.Name Type

data Type
    = TyParam Name
    | TyVariable TypeVariable
    | TyConstructor { tyConstructor :: Name, tyArguments :: [Type] }
    | TyFunction [Type] Type
    deriving (Show, Eq, Generic, Typeable)

pattern TyInteger, TyBoolean :: Type
pattern TyInteger = TyConstructor "Integer" []
pattern TyBoolean = TyConstructor "Boolean" []

type TypeScheme = U.Bind [TypeVariable] Type

scheme :: Type -> TypeScheme
scheme = U.bind []

type TermVariable = U.Name Term

data Term
    = Literal Literal
    | Binary Term Operation Term
    | Variable TermVariable
    | Lambda (U.Bind [Parameter] Term)
    | Application Term [Type] [Term]
    deriving (Show, Generic, Typeable)

data Literal = LitInteger Int | LitBoolean Bool deriving (Show, Generic, Typeable)
data Operation = Plus | Minus | And | Or | Equals deriving (Show, Generic, Typeable)
data Parameter = Parameter { paramName :: TermVariable, paramType :: Type } deriving (Show, Generic, Typeable)

-- Names and binders using unbound-generics
-- https://hackage.haskell.org/package/unbound-generics
-- https://hackage.haskell.org/package/unbound-generics-unify

instance U.Alpha Type
instance U.Subst Type Type where
  isvar (TyVariable var) = Just (U.SubstName var)
  isvar _ = Nothing
instance U.Unify Type Type

instance U.Alpha Literal
instance U.Alpha Operation
instance U.Alpha Parameter
instance U.Alpha Term
