{-# LANGUAGE MultiWayIf, ApplicativeDo #-}
module TypesWorkshop.Simple.Type where

import TypesWorkshop.Simple.Term
import TypesWorkshop.ValidationNel

data Type = TyInteger | TyBoolean deriving (Show, Eq)

data TypeError = TypeError { expected, actual :: Type } deriving (Show)

shouldBe :: Type -> Type -> ValidationNel TypeError ()
shouldBe actual expected
  | actual == expected = pure ()
  | otherwise          = problem $ TypeError expected actual

sameTypeOperation :: Type -> Type -> Type -> ValidationNel TypeError Type
sameTypeOperation tyLeft tyRight expected = do
    tyLeft `shouldBe` expected
    tyRight `shouldBe` expected
    pure expected

typeCheck :: Term -> ValidationNel TypeError Type
typeCheck (Literal (LitInteger _)) = pure TyInteger
typeCheck (Literal (LitBoolean _)) = pure TyBoolean
typeCheck (Binary left op right) = do
    tyLeft <- typeCheck left
    tyRight <- typeCheck right
    case op of
        Plus  -> sameTypeOperation tyLeft tyRight TyInteger
        Minus -> sameTypeOperation tyLeft tyRight TyInteger
        And   -> sameTypeOperation tyLeft tyRight TyBoolean
        Or    -> sameTypeOperation tyLeft tyRight TyBoolean
        Equals -> do tyLeft `shouldBe` tyRight
                     pure TyBoolean

-- >>> typeCheck (Binary (Literal (LitInteger 1)) Plus (Literal (LitInteger 2)))
-- Success TyInteger

-- >>> typeCheck (Binary (Literal (LitInteger 1)) Plus (Literal (LitBoolean True)))
-- Failure (TypeError {expected = TyInteger, actual = TyBoolean} :| [])

-- >>> let p = Binary (Literal (LitInteger 1)) Plus (Literal (LitBoolean True))
-- >>> let t = Binary p Equals p
-- >>> typeCheck t
-- Failure (TypeError {expected = TyInteger, actual = TyBoolean} :| [TypeError {expected = TyInteger, actual = TyBoolean}])
