package com.serranofp.types.context

import me.alllex.parsus.parser.*
import me.alllex.parsus.token.*

object TermParser : Grammar<Term>() {
    init { regexToken("\\s+", ignored = true) }

    val lpar by literalToken("(")
    val rpar by literalToken(")")
    val comma by literalToken(",")
    val lam by literalToken("fun")
    val colon by literalToken(":")
    val arrow by literalToken("->")

    val tru by literalToken("true") map { Literal.True }
    val fal by literalToken("false") map { Literal.False }

    val plus by literalToken("+") map { Operation.Plus }
    val minus by literalToken("-") map { Operation.Minus }
    val and by literalToken("&&") map { Operation.And }
    val or by literalToken("||") map { Operation.Or }
    val equals by literalToken("==") map { Operation.Equals }

    val Integer by literalToken("Integer") map { Type.Integer }
    val Boolean by literalToken("Boolean") map { Type.Boolean }
    val FunctionArguments by parser {
        lpar()
        val args = separated(type, comma, allowEmpty = false)()
        rpar()
        args
    }
    val Function by parser {
        val argumentTypes = FunctionArguments()
        arrow()
        val returnType = type()
        Type.Function(argumentTypes, returnType)
    }
    val type: Parser<Type> by Integer or Boolean or Function

    val identifier by regexToken("\\w+") map { it.text }
    val variable by identifier map { Variable(it) }

    val parameter by identifier and -colon and type map { (n, t) -> Parameter(n, t) }
    val parameters by -lpar * separated(parameter, comma, allowEmpty = false) * -rpar

    val lambda by parser {
        lam()
        val params = parameters()
        arrow()
        val body = expr()
        Lambda(params, body)
    }

    val number by regexToken("\\d+") map { Literal.Integer(it.text.toInt()) }
    val literal by tru or fal or number

    val parenthesized by parser {
        lpar()
        val e = expr()
        rpar()
        e
    }

    val atom: Parser<Term> by lambda or literal or variable or parenthesized

    val arithExpr by parser<Term> {
        reduce(atom, plus or minus) { l, o, r -> Binary(l, o, r) }
    }
    val addExpr by parser<Term> {
        reduce(arithExpr, and) { l, o, r -> Binary(l, o, r) }
    }
    val orExpr by parser<Term> {
        reduce(addExpr, or) { l, o, r -> Binary(l, o, r) }
    }
    val equalExpr by parser<Term> {
        reduce(orExpr, equals) { l, o, r -> Binary(l, o, r) }
    }
    val expr by equalExpr

    override val root by expr
}