package com.serranofp.types.generics

import arrow.core.EitherNel
import arrow.core.raise.either
import arrow.core.raise.mapOrAccumulate

data class Program(
    val declarations: List<Declaration>
)

data class Declaration(
    val name: Name,
    val type: TypeScheme,
    val body: Term?
)

fun Program.resolutionCheck(): EitherNel<ResolutionError, Program> = either {
    val valueContext = declarations.map { it.name }
    Program(mapOrAccumulate(declarations) { it.resolutionCheck(valueContext).bind() })
}

fun Declaration.resolutionCheck(valueContext: List<Name>): EitherNel<ResolutionError, Declaration> = either {
    Declaration(
        name,
        type.resolutionCheck().bind(),
        body?.resolutionCheck(type.typeParameters, valueContext)?.bind()
    )
}
