package com.serranofp.types.simple

import me.alllex.parsus.parser.*
import me.alllex.parsus.token.*

object TermParser : Grammar<Term>() {
    init { regexToken("\\s+", ignored = true) }

    val lpar by literalToken("(")
    val rpar by literalToken(")")

    val tru by literalToken("true") map { Literal.True }
    val fal by literalToken("false") map { Literal.False }

    val plus by literalToken("+") map { Operation.Plus }
    val minus by literalToken("-") map { Operation.Minus }
    val and by literalToken("&&") map { Operation.And }
    val or by literalToken("||") map { Operation.Or }
    val equals by literalToken("==") map { Operation.Equals }

    val number by regexToken("\\d+") map { Literal.Integer(it.text.toInt()) }

    val literal by tru or fal or number
    val parenthesized by parser { skip(lpar) * root() * skip(rpar) }
    val atom: Parser<Term> by literal or parenthesized

    val arithExpr by parser<Term> {
        reduce(atom, plus or minus) { l, o, r -> Binary(l, o, r) }
    }
    val addExpr by parser<Term> {
        reduce(arithExpr, and) { l, o, r -> Binary(l, o, r) }
    }
    val orExpr by parser<Term> {
        reduce(addExpr, or) { l, o, r -> Binary(l, o, r) }
    }
    val equalExpr by parser<Term> {
        reduce(orExpr, equals) { l, o, r -> Binary(l, o, r) }
    }

    override val root by equalExpr
}