package com.serranofp.types.simple

import arrow.core.Either
import arrow.core.raise.either
import arrow.core.raise.ensure

sealed interface Type {
    data object Integer : Type
    data object Boolean : Type
}

data class TypeError(val expected: Type, val actual: Type)

fun Term.typeCheck(): Either<TypeError, Type> = either {
    // utility function to check types
    infix fun Type.shouldBe(other: Type) {
        ensure(this == other) { TypeError(other, this) }
    }

    when (this@typeCheck) {
        is Literal.Integer -> Type.Integer
        Literal.True, Literal.False -> Type.Boolean
        is Binary -> {
            val leftType = left.typeCheck().bind()
            val rightType = right.typeCheck().bind()
            when (operation) {
                Operation.Plus, Operation.Minus -> {
                    leftType shouldBe Type.Integer
                    rightType shouldBe Type.Integer
                    Type.Integer
                }
                Operation.And, Operation.Or -> {
                    leftType shouldBe Type.Boolean
                    rightType shouldBe Type.Boolean
                    Type.Boolean
                }
                Operation.Equals -> {
                    rightType shouldBe leftType
                    Type.Boolean
                }
            }
        }
    }
}
