package com.serranofp.types.unification

import arrow.core.Either
import arrow.core.left
import arrow.core.right

interface BasicTerm: UnifiableTerm<String, String, BasicTerm>

open class BasicVariable(override val variable: String)
    : UnifiableVariable<String, String, BasicTerm>(), BasicTerm

fun String.v(): BasicVariable = BasicVariable(this)

open class BasicFunctor(override val head: String, override val subterms: List<BasicTerm>)
    : UnifiableFunctor<String, String, BasicTerm>(), BasicTerm {
    override fun substitute(substitution: Substitution<String, BasicTerm>): BasicTerm =
        BasicFunctor(head, subterms.map { it.substitute(substitution) })
}

fun String.t(vararg args: BasicTerm) = BasicFunctor(this, args.asList())

fun unifyClean(a: BasicTerm, b: BasicTerm): Either<UnifierStatus, Substitution<String, BasicTerm>> = Unifier<String, String, BasicTerm>().run {
    unify(a, b)
    if (status == UnifierStatus.VALID) substitution.right() else status.left()
}

fun main() {
    val t1 = "f".t("a".t(), "b".t())
    val t2 = "f".t("x".v(), "y".v())
    val t3 = "f".t("x".v(), "x".v())
    val t4 = "f".t("g".t("x".v()), "x".v())
    val t5 = "g".t("a".t(), "b".t())
    val t6 = "g".t("a".t(), "b".t(), "x".t())
    println(unifyClean(t1, t1))
    println(unifyClean(t1, t2))
    println(unifyClean(t2, t1))
    println(unifyClean(t1, t3))
    println(unifyClean(t2, t3))
    println(unifyClean(t2, t4))
    println(unifyClean(t1, t5))
    println(unifyClean(t2, t5))
    println(unifyClean(t5, t6))
}